package org.eclipse.jdt.internal.compiler.classfmt;

package .tmp;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.classfmt.AnnotationInfo;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFileStruct;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFormatException;
import org.eclipse.jdt.internal.compiler.classfmt.FieldInfoWithAnnotation;
import org.eclipse.jdt.internal.compiler.classfmt.FieldInfoWithTypeAnnotation;
import org.eclipse.jdt.internal.compiler.classfmt.TypeAnnotationInfo;
import org.eclipse.jdt.internal.compiler.codegen.AttributeNamesConstants;
import org.eclipse.jdt.internal.compiler.env.IBinaryAnnotation;
import org.eclipse.jdt.internal.compiler.env.IBinaryField;
import org.eclipse.jdt.internal.compiler.env.IBinaryTypeAnnotation;
import org.eclipse.jdt.internal.compiler.impl.BooleanConstant;
import org.eclipse.jdt.internal.compiler.impl.ByteConstant;
import org.eclipse.jdt.internal.compiler.impl.CharConstant;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.impl.DoubleConstant;
import org.eclipse.jdt.internal.compiler.impl.FloatConstant;
import org.eclipse.jdt.internal.compiler.impl.IntConstant;
import org.eclipse.jdt.internal.compiler.impl.LongConstant;
import org.eclipse.jdt.internal.compiler.impl.ShortConstant;
import org.eclipse.jdt.internal.compiler.impl.StringConstant;
import org.eclipse.jdt.internal.compiler.util.Util;

public class FieldInfo extends ClassFileStruct implements IBinaryField, Comparable {
  protected int accessFlags;
  
  protected int attributeBytes;
  
  protected Constant constant;
  
  protected char[] descriptor;
  
  protected char[] name;
  
  protected char[] signature;
  
  protected int signatureUtf8Offset;
  
  protected long tagBits;
  
  protected Object wrappedConstantValue;
  
  public static org.eclipse.jdt.internal.compiler.classfmt.FieldInfo createField(byte[] classFileBytes, int[] offsets, int offset) {
    org.eclipse.jdt.internal.compiler.classfmt.FieldInfo fieldInfo = new org.eclipse.jdt.internal.compiler.classfmt.FieldInfo(classFileBytes, offsets, offset);
    int attributesCount = fieldInfo.u2At(6);
    int readOffset = 8;
    AnnotationInfo[] annotations = null;
    TypeAnnotationInfo[] typeAnnotations = null;
    for (int i = 0; i < attributesCount; i++) {
      int utf8Offset = fieldInfo.constantPoolOffsets[fieldInfo.u2At(readOffset)] - fieldInfo.structOffset;
      char[] attributeName = fieldInfo.utf8At(utf8Offset + 3, fieldInfo.u2At(utf8Offset + 1));
      if (attributeName.length > 0) {
        AnnotationInfo[] decodedAnnotations;
        TypeAnnotationInfo[] decodedTypeAnnotations;
        switch (attributeName[0]) {
          case 'S':
            if (CharOperation.equals(AttributeNamesConstants.SignatureName, attributeName))
              fieldInfo.signatureUtf8Offset = fieldInfo.constantPoolOffsets[fieldInfo.u2At(readOffset + 6)] - fieldInfo.structOffset; 
            break;
          case 'R':
            decodedAnnotations = null;
            decodedTypeAnnotations = null;
            if (CharOperation.equals(attributeName, AttributeNamesConstants.RuntimeVisibleAnnotationsName)) {
              decodedAnnotations = fieldInfo.decodeAnnotations(readOffset, true);
            } else if (CharOperation.equals(attributeName, AttributeNamesConstants.RuntimeInvisibleAnnotationsName)) {
              decodedAnnotations = fieldInfo.decodeAnnotations(readOffset, false);
            } else if (CharOperation.equals(attributeName, AttributeNamesConstants.RuntimeVisibleTypeAnnotationsName)) {
              decodedTypeAnnotations = fieldInfo.decodeTypeAnnotations(readOffset, true);
            } else if (CharOperation.equals(attributeName, AttributeNamesConstants.RuntimeInvisibleTypeAnnotationsName)) {
              decodedTypeAnnotations = fieldInfo.decodeTypeAnnotations(readOffset, false);
            } 
            if (decodedAnnotations != null) {
              if (annotations == null) {
                annotations = decodedAnnotations;
                break;
              } 
              int length = annotations.length;
              AnnotationInfo[] combined = new AnnotationInfo[length + decodedAnnotations.length];
              System.arraycopy(annotations, 0, combined, 0, length);
              System.arraycopy(decodedAnnotations, 0, combined, length, decodedAnnotations.length);
              annotations = combined;
              break;
            } 
            if (decodedTypeAnnotations != null) {
              if (typeAnnotations == null) {
                typeAnnotations = decodedTypeAnnotations;
                break;
              } 
              int length = typeAnnotations.length;
              TypeAnnotationInfo[] combined = new TypeAnnotationInfo[length + decodedTypeAnnotations.length];
              System.arraycopy(typeAnnotations, 0, combined, 0, length);
              System.arraycopy(decodedTypeAnnotations, 0, combined, length, decodedTypeAnnotations.length);
              typeAnnotations = combined;
            } 
            break;
        } 
      } 
      readOffset = (int)(readOffset + 6L + fieldInfo.u4At(readOffset + 2));
    } 
    fieldInfo.attributeBytes = readOffset;
    if (typeAnnotations != null)
      return (org.eclipse.jdt.internal.compiler.classfmt.FieldInfo)new FieldInfoWithTypeAnnotation(fieldInfo, annotations, typeAnnotations); 
    if (annotations != null)
      return (org.eclipse.jdt.internal.compiler.classfmt.FieldInfo)new FieldInfoWithAnnotation(fieldInfo, annotations); 
    return fieldInfo;
  }
  
  protected FieldInfo(byte[] classFileBytes, int[] offsets, int offset) {
    super(classFileBytes, offsets, offset);
    this.accessFlags = -1;
    this.signatureUtf8Offset = -1;
  }
  
  private AnnotationInfo[] decodeAnnotations(int offset, boolean runtimeVisible) {
    int numberOfAnnotations = u2At(offset + 6);
    if (numberOfAnnotations > 0) {
      int readOffset = offset + 8;
      AnnotationInfo[] newInfos = null;
      int newInfoCount = 0;
      for (int i = 0; i < numberOfAnnotations; i++) {
        AnnotationInfo newInfo = new AnnotationInfo(this.reference, this.constantPoolOffsets, 
            readOffset + this.structOffset, runtimeVisible, false);
        readOffset += newInfo.readOffset;
        long standardTagBits = newInfo.standardAnnotationTagBits;
        if (standardTagBits != 0L) {
          this.tagBits |= standardTagBits;
        } else {
          if (newInfos == null)
            newInfos = new AnnotationInfo[numberOfAnnotations - i]; 
          newInfos[newInfoCount++] = newInfo;
        } 
      } 
      if (newInfos != null) {
        if (newInfoCount != newInfos.length)
          System.arraycopy(newInfos, 0, newInfos = new AnnotationInfo[newInfoCount], 0, newInfoCount); 
        return newInfos;
      } 
    } 
    return null;
  }
  
  TypeAnnotationInfo[] decodeTypeAnnotations(int offset, boolean runtimeVisible) {
    int numberOfAnnotations = u2At(offset + 6);
    if (numberOfAnnotations > 0) {
      int readOffset = offset + 8;
      TypeAnnotationInfo[] typeAnnos = new TypeAnnotationInfo[numberOfAnnotations];
      for (int i = 0; i < numberOfAnnotations; i++) {
        TypeAnnotationInfo newInfo = new TypeAnnotationInfo(this.reference, this.constantPoolOffsets, readOffset + this.structOffset, runtimeVisible, false);
        readOffset += newInfo.readOffset;
        typeAnnos[i] = newInfo;
      } 
      return typeAnnos;
    } 
    return null;
  }
  
  public int compareTo(Object o) {
    return (new String(getName())).compareTo(new String(((org.eclipse.jdt.internal.compiler.classfmt.FieldInfo)o).getName()));
  }
  
  public boolean equals(Object o) {
    if (!(o instanceof org.eclipse.jdt.internal.compiler.classfmt.FieldInfo))
      return false; 
    return CharOperation.equals(getName(), ((org.eclipse.jdt.internal.compiler.classfmt.FieldInfo)o).getName());
  }
  
  public int hashCode() {
    return CharOperation.hashCode(getName());
  }
  
  public Constant getConstant() {
    if (this.constant == null)
      readConstantAttribute(); 
    return this.constant;
  }
  
  public char[] getGenericSignature() {
    if (this.signatureUtf8Offset != -1) {
      if (this.signature == null)
        this.signature = utf8At(this.signatureUtf8Offset + 3, u2At(this.signatureUtf8Offset + 1)); 
      return this.signature;
    } 
    return null;
  }
  
  public int getModifiers() {
    if (this.accessFlags == -1) {
      this.accessFlags = u2At(0);
      readModifierRelatedAttributes();
    } 
    return this.accessFlags;
  }
  
  public char[] getName() {
    if (this.name == null) {
      int utf8Offset = this.constantPoolOffsets[u2At(2)] - this.structOffset;
      this.name = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
    } 
    return this.name;
  }
  
  public long getTagBits() {
    return this.tagBits;
  }
  
  public char[] getTypeName() {
    if (this.descriptor == null) {
      int utf8Offset = this.constantPoolOffsets[u2At(4)] - this.structOffset;
      this.descriptor = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
    } 
    return this.descriptor;
  }
  
  public IBinaryAnnotation[] getAnnotations() {
    return null;
  }
  
  public IBinaryTypeAnnotation[] getTypeAnnotations() {
    return null;
  }
  
  public Object getWrappedConstantValue() {
    if (this.wrappedConstantValue == null && 
      hasConstant()) {
      Constant fieldConstant = getConstant();
      switch (fieldConstant.typeID()) {
        case 10:
          this.wrappedConstantValue = Integer.valueOf(fieldConstant.intValue());
          break;
        case 3:
          this.wrappedConstantValue = Byte.valueOf(fieldConstant.byteValue());
          break;
        case 4:
          this.wrappedConstantValue = Short.valueOf(fieldConstant.shortValue());
          break;
        case 2:
          this.wrappedConstantValue = Character.valueOf(fieldConstant.charValue());
          break;
        case 9:
          this.wrappedConstantValue = new Float(fieldConstant.floatValue());
          break;
        case 8:
          this.wrappedConstantValue = new Double(fieldConstant.doubleValue());
          break;
        case 5:
          this.wrappedConstantValue = Util.toBoolean(fieldConstant.booleanValue());
          break;
        case 7:
          this.wrappedConstantValue = Long.valueOf(fieldConstant.longValue());
          break;
        case 11:
          this.wrappedConstantValue = fieldConstant.stringValue();
          break;
      } 
    } 
    return this.wrappedConstantValue;
  }
  
  public boolean hasConstant() {
    return (getConstant() != Constant.NotAConstant);
  }
  
  protected void initialize() {
    getModifiers();
    getName();
    getConstant();
    getTypeName();
    getGenericSignature();
    reset();
  }
  
  public boolean isSynthetic() {
    return ((getModifiers() & 0x1000) != 0);
  }
  
  private void readConstantAttribute() {
    int attributesCount = u2At(6);
    int readOffset = 8;
    boolean isConstant = false;
    for (int i = 0; i < attributesCount; i++) {
      int utf8Offset = this.constantPoolOffsets[u2At(readOffset)] - this.structOffset;
      char[] attributeName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
      if (CharOperation.equals(attributeName, AttributeNamesConstants.ConstantValueName)) {
        char[] sign;
        isConstant = true;
        int relativeOffset = this.constantPoolOffsets[u2At(readOffset + 6)] - this.structOffset;
        switch (u1At(relativeOffset)) {
          case 3:
            sign = getTypeName();
            if (sign.length == 1) {
              switch (sign[0]) {
                case 'Z':
                  this.constant = BooleanConstant.fromValue((i4At(relativeOffset + 1) == 1));
                  break;
                case 'I':
                  this.constant = IntConstant.fromValue(i4At(relativeOffset + 1));
                  break;
                case 'C':
                  this.constant = CharConstant.fromValue((char)i4At(relativeOffset + 1));
                  break;
                case 'B':
                  this.constant = ByteConstant.fromValue((byte)i4At(relativeOffset + 1));
                  break;
                case 'S':
                  this.constant = ShortConstant.fromValue((short)i4At(relativeOffset + 1));
                  break;
              } 
              this.constant = Constant.NotAConstant;
              break;
            } 
            this.constant = Constant.NotAConstant;
            break;
          case 4:
            this.constant = FloatConstant.fromValue(floatAt(relativeOffset + 1));
            break;
          case 6:
            this.constant = DoubleConstant.fromValue(doubleAt(relativeOffset + 1));
            break;
          case 5:
            this.constant = LongConstant.fromValue(i8At(relativeOffset + 1));
            break;
          case 8:
            utf8Offset = this.constantPoolOffsets[u2At(relativeOffset + 1)] - this.structOffset;
            this.constant = 
              StringConstant.fromValue(
                String.valueOf(utf8At(utf8Offset + 3, u2At(utf8Offset + 1))));
            break;
        } 
      } 
      readOffset = (int)(readOffset + 6L + u4At(readOffset + 2));
    } 
    if (!isConstant)
      this.constant = Constant.NotAConstant; 
  }
  
  private void readModifierRelatedAttributes() {
    int attributesCount = u2At(6);
    int readOffset = 8;
    for (int i = 0; i < attributesCount; i++) {
      int utf8Offset = this.constantPoolOffsets[u2At(readOffset)] - this.structOffset;
      char[] attributeName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
      if (attributeName.length != 0)
        switch (attributeName[0]) {
          case 'D':
            if (CharOperation.equals(attributeName, AttributeNamesConstants.DeprecatedName))
              this.accessFlags |= 0x100000; 
            break;
          case 'S':
            if (CharOperation.equals(attributeName, AttributeNamesConstants.SyntheticName))
              this.accessFlags |= 0x1000; 
            break;
        }  
      readOffset = (int)(readOffset + 6L + u4At(readOffset + 2));
    } 
  }
  
  public int sizeInBytes() {
    return this.attributeBytes;
  }
  
  public void throwFormatException() throws ClassFormatException {
    throw new ClassFormatException(17);
  }
  
  public String toString() {
    StringBuffer buffer = new StringBuffer(getClass().getName());
    toStringContent(buffer);
    return buffer.toString();
  }
  
  protected void toStringContent(StringBuffer buffer) {
    int modifiers = getModifiers();
    buffer
      .append('{')
      .append(
        String.valueOf(((modifiers & 0x100000) != 0) ? "deprecated " : Util.EMPTY_STRING) + (
        ((modifiers & 0x1) == 1) ? "public " : Util.EMPTY_STRING) + (
        ((modifiers & 0x2) == 2) ? "private " : Util.EMPTY_STRING) + (
        ((modifiers & 0x4) == 4) ? "protected " : Util.EMPTY_STRING) + (
        ((modifiers & 0x8) == 8) ? "static " : Util.EMPTY_STRING) + (
        ((modifiers & 0x10) == 16) ? "final " : Util.EMPTY_STRING) + (
        ((modifiers & 0x40) == 64) ? "volatile " : Util.EMPTY_STRING) + (
        ((modifiers & 0x80) == 128) ? "transient " : Util.EMPTY_STRING))
      .append(getTypeName())
      .append(' ')
      .append(getName())
      .append(' ')
      .append(getConstant())
      .append('}')
      .toString();
  }
}
